/*
	$Id: AbstractSqlQuery.java 6880 2015-04-21 06:26:57Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.hibernate.ResultSetWork;
import net.sf.basedb.core.hibernate.SavePointWork;
import net.sf.basedb.core.query.QueryParameter;
import net.sf.basedb.core.query.SqlQuery;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.signal.SignalException;

import java.util.List;
import java.util.LinkedList;
import java.util.Collection;
import java.util.regex.Pattern;
import java.util.regex.Matcher;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Savepoint;
import java.sql.Statement;

import org.hibernate.dialect.pagination.LimitHandler;
import org.hibernate.engine.spi.RowSelection;

/**
	An abstract implementation of the {@link SqlQuery} interface. 

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-21 08:26:57 +0200 (ti, 21 apr 2015) $
*/
public abstract class AbstractSqlQuery
	extends AbstractQuery
	implements SqlQuery
{
	/**
		Log all SQL statements.
	*/
	private static final org.slf4j.Logger logSql = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.query.ql");

	/**
		So we don't always have to call logSql.debug()
	*/
	private static final boolean debugSqlEnabled = logSql.isDebugEnabled();

	/**
		Log all parameter bindings to prepared statement.
	*/
	private static final org.slf4j.Logger logParam = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.query.bind");
	
	/**
		So we don't always have to call logParam.debug()
	*/
	private static final boolean debugParamEnabled = logParam.isDebugEnabled();
		
	/**
		Pattern that finds a named parameter in an SQL string. A named parameter
		is a colon (:) followed by one or more alphanumeric characters.
	*/
	private static final Pattern namedParameter = Pattern.compile(":(\\w+)");

	/**
	*/
	AbstractSqlQuery(String rootEntity)
	{
		super(rootEntity);
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		@return {@link QueryType#SQL}
	*/
	@Override
	public QueryType getQueryType()
	{
		return QueryType.SQL;
	}
	@Override
	public long count(DbControl dc)
		throws BaseException
	{
		long totalCount = -1;
		CountWork work = createCountWork(dc);
		try
		{
			if (debugSqlEnabled) logSql.debug("Executing count query: " + work.getSql());
			boolean useSavePoint = isFailSafe() && 
					HibernateUtil.getDbEngine().useSavePointToContinueTransactionFromSqlFailure();
			totalCount = HibernateUtil.doReturningWork(dc.getHibernateSession(), 
				useSavePoint ? new SavePointWork<Long>(work) : work);
		}
		catch (SQLException ex)
		{
			logSql.error(work.getSql(), ex);
			throw new BaseException(ex);
		}
		return totalCount;
	}
	// -------------------------------------------
	
	/**
		Creates a work item for counting the number of total hits by the query.
	*/
	private CountWork createCountWork(DbControl dc)
	{
		String countSql = getCountQuery(dc, true);
		List<String> countParameterOrder = new LinkedList<String>();
		countSql = parseParameters(countSql, countParameterOrder);
		return new CountWork(dc.getSessionControl(), countSql, countParameterOrder);
	}
	
	abstract Select[] getDefaultSelects();

	/**
		Execute the query.
		@param dc The <code>DbControl</code> to use for database access
		@return A <code>DynamicResultIterator</code> object
		@throws BaseException If there is an error
	*/
	@Override
	public DynamicResultIterator iterate(DbControl dc)
		throws BaseException
	{
		long totalCount = -1;

		// Reset temporary query elements
		resetTemporary();

		// If no selections has been specified, we use the defaults...
		if ((getSelects() == null || getSelects().size() == 0) && 
			(getPermanentSelects() == null || getPermanentSelects().size() == 0))
		{
			Select[] defaults = getDefaultSelects();
			if (defaults != null)
			{
				for (Select s : defaults)
				{
					if (s != null) temporarySelect(s);
				}
			}
		}
		
		// Get the main query
		String mainSql = getMainQuery(dc, true);
		
		// Holds the names of the parameters in the query in the order they appear
		List<String> parameterOrder = new LinkedList<String>();
		
		// Parse out named parameters
		mainSql = parseParameters(mainSql, parameterOrder);
		
		// Add limit parameters
		LimitHandler limitHandler = getLimitHandler(mainSql);
		if (limitHandler != null)
		{
			mainSql = limitHandler.getProcessedSql();
		}
		// Main query is now built
		
		ResultSet result = null;
		Connection conn = null;
		Savepoint savepoint = null;
		PreparedStatement ps = null;
		RuntimeException exception = null;
		String currentSql = null;
		try
		{
			conn = HibernateUtil.getConnection(dc.getHibernateSession());		
			if (isFailSafe() && 
				HibernateUtil.getDbEngine().useSavePointToContinueTransactionFromSqlFailure())
			{
				savepoint = conn.setSavepoint();
			}

			// Load the total count if it is requested
			if (isReturningTotalCount())
			{
				CountWork countWork = createCountWork(dc);
				currentSql = countWork.getSql();
				if (debugSqlEnabled) logSql.debug("Executing count query: " + currentSql);
				totalCount = countWork.execute(conn);
			}
			
			if (debugSqlEnabled) logSql.debug("Executing main query: " + mainSql);
			currentSql = mainSql;
			ps = conn.prepareStatement(mainSql);
			int parameterIndex = 1;
			if (limitHandler != null)
			{
				parameterIndex += limitHandler.bindLimitParametersAtStartOfQuery(ps, parameterIndex);
			}
			parameterIndex += setParameters(ps, parameterOrder, parameterIndex);
			if (limitHandler != null)
			{
				parameterIndex += limitHandler.bindLimitParametersAtEndOfQuery(ps, parameterIndex);
			}
			
			result = QueryExecutor.executeQuery(ps, dc.getSessionControl());
			if (limitHandler != null && !limitHandler.supportsLimitOffset())
			{
				result.absolute(getFirstResult());
			}
		}
		catch (InterruptedException ex)
		{
			exception = new SignalException("Aborted by user.", ex);
		}
		catch (SQLException ex)
		{
			logSql.error(currentSql, ex);
			exception = new BaseException(ex);
		}
		finally
		{
			if (exception != null)
			{
				if (result != null) closeResultSet(result, currentSql);
				if (ps != null) closeStatement(ps, currentSql);
				if (savepoint != null) rollbackToSavepoint(conn, savepoint, currentSql);
				throw exception;
			}
			else
			{
				if (savepoint != null) releaseSavepoint(conn, savepoint, currentSql);
			}
		}
		return new DynamicResultIterator(result, totalCount, getMaxResults());
	}

	/**
		Parse the given SQL and find named parameters. A named parameter
		is a colon (:) followed by one or more alphanumeric characters.
		Replaces the named parameters with a JDBC parameter placeholder,
		ie. a question mark. The name of the parameters are put in the
		specified list in the order they appear in the SQL string.
		@param sql The original SQL string, which may contain named
			parameters
		@param parameterOrder A list that should store the name of the
			parameters in the correct order
		@return The new SQL statement with JDBC question marks instead
			of named parameters
	*/
	String parseParameters(String sql, List<String> parameterOrder)
	{
		Matcher m = namedParameter.matcher(sql);
		StringBuffer sb = new StringBuffer();
		while (m.find())
		{
			String name = m.group(1);
			if (parameterOrder != null) parameterOrder.add(name);
			Object parameterValue = getParameterValue(name);
			if (parameterValue instanceof Collection)
			{
				// If the parameter is a collection one question mark is required for each value
				int size = ((Collection<?>)parameterValue).size();
				StringBuilder sb2 = new StringBuilder();
				for (int i = 0; i < size; ++i)
				{
					if (i > 0) sb2.append(",");
					sb2.append("?");
				}
				m.appendReplacement(sb, sb2.toString());
			}
			else
			{
				m.appendReplacement(sb, "?");
			}
		}
		m.appendTail(sb);
		return sb.toString();
	}

	/**
		Parse the given SQL and add LIMIT and OFFSET parameters to it. This method
		checks the capabilities of the installed Hibernate <code>Dialect</code>
		and uses it for generating the actual SQL. If limit parameters are added
		to the query this method return a LimitHandler object. The caller
		must get the resulting SQL by calling LimitHandler.getProcessedSql().
		and then use the LimitHandler.bindbindLimitParameters*() methods
		to actually bind the limit parameters.	
		<p>
		If the underlying database doesn't support limit parameters the query
		will be executed and the result starting point moved forward to the
		correct offset and the a limit will be passed to the {@link DynamicResultIterator}
		object.

		@param sql The original SQL without limit parameters
		@return The new SQL with limit parameters
	*/
	LimitHandler getLimitHandler(String sql)
	{
		LimitHandler limitHandler = null;
		org.hibernate.dialect.Dialect dialect = HibernateUtil.getDialect();
		boolean hasLimit = getMaxResults() > 0;
		boolean hasOffset = getFirstResult() > 0;
		if ((hasLimit || hasOffset))
		{
			RowSelection rowSelection = new RowSelection();
			if (hasLimit) rowSelection.setMaxRows(getMaxResults());
			if (hasOffset) rowSelection.setFirstRow(getFirstResult());
			
			limitHandler = dialect.buildLimitHandler(sql, rowSelection);
		}
		return limitHandler;
	}

	/**
		Set parameter values on a prepared statement.
		@param ps The prepared statement
		@param parameterOrder The names of the parameters in the order they
			appear in the prepared statement
		@param parameterIndex The index of the first parameter to bind
		@return The number of parameters bound
	*/
	int setParameters(PreparedStatement ps, List<String> parameterOrder, int parameterIndex)
		throws SQLException
	{
		int i = parameterIndex;
		for (String name : parameterOrder)
		{
			QueryParameter qp = getQueryParameter(name);
			Object parameterValue = null;
			Type valueType = null;
			if (qp != null) 
			{
				qp.validateValue();
				parameterValue = qp.getValue();
				valueType = qp.getType();
			}
			if (parameterValue instanceof Collection)
			{
				if (debugParamEnabled) logParam.debug("Binding collection to parameter index: "+i+"++");
				for (Object value : (Collection<?>)parameterValue)
				{
					setParameterValue(ps, i++, value, valueType);
				}
			}
			else
			{
				setParameterValue(ps, i++, parameterValue, valueType);
			}
		}
		return i - parameterIndex;
	}
	
	/**
		Set a parameter value on a prepared statement. If the object is a
		{@link BasicItem} the id of the object is used a value.

		@param ps The prepared statement
		@param index The index of the parameter, starting at 1
		@param parameterValue The value of the parameter
	*/
	private void setParameterValue(PreparedStatement ps, int index, Object parameterValue, Type valueType)
		throws SQLException
	{
		if (parameterValue instanceof BasicItem)
		{
			int id = ((BasicItem)parameterValue).getId();
			if (debugParamEnabled) logParam.debug("Binding '"+id+"' to parameter index: "+index);
			ps.setInt(index, id);
		}
		else
		{
			if (debugParamEnabled) logParam.debug("Binding '"+parameterValue+"' to parameter index: "+index);
			if (valueType == null)
			{
				ps.setObject(index, parameterValue);
			}
			else
			{
				ps.setObject(index, parameterValue, valueType.getSQLType());
			}
		}
	}
	
	/**
		Close the result set without throwing an exception. If
		the {@link ResultSet#close()} method throws an exception
		it is logged together with the message.
		@param msg A message to write to the log if the close fails
		@param result The result set to close
		@since 3.1
	*/
	protected void closeResultSet(ResultSet result, String msg)
	{
		try
		{
			result.close();
		}
		catch (Throwable t)
		{
			logSql.warn(msg, t);
		}
	}
	
	/**
		Close the statement without throwing an exception. If
		the {@link Statement#close()} method throws an exception
		it is logged together with the message.
		@param msg A message to write to the log if the close fails
		@param stmt The statement to close
		@since 3.1
	*/
	protected void closeStatement(Statement stmt, String msg)
	{
		try
		{
			stmt.close();
		}
		catch (Throwable t)
		{
			logSql.warn(msg, t);
		}
	}
	
	/**
		Rollback the current transaction to the given savepoint.
		@param conn The connection to rollback
		@param savepoint The savepoint to rollback to
		@param msg A message to write to the log if the rollback fails
		@since 3.1
	*/
	protected void rollbackToSavepoint(Connection conn, Savepoint savepoint, String msg)
	{
		try
		{
			conn.rollback(savepoint);
		}
		catch (Throwable t)
		{
			logSql.warn(msg, t);
		}
	}

	/**
		Release the given savepoint.
		@param conn The connection that holds the savepoint
		@param savepoint The savepoint to release
		@param msg A message to write to the log if the release fails
		@since 3.1
	*/
	protected void releaseSavepoint(Connection conn, Savepoint savepoint, String msg)
	{
		try
		{
			conn.releaseSavepoint(savepoint);
		}
		catch (Throwable t)
		{
			logSql.warn(msg, t);
		}
	}
	
	
	class CountWork
		extends ResultSetWork<Long>
	{

		
		private List<String> parameterOrder;
		
		CountWork(SessionControl sc, String sql, List<String> parameterOrder)
		{
			super(sc, sql);
			this.parameterOrder = parameterOrder;
		}


		@Override
		protected void setParameters(PreparedStatement ps)
			throws SQLException
		{
			AbstractSqlQuery.this.setParameters(ps, parameterOrder, 1);
		}

		@Override
		protected Long getResult(ResultSet rs)
			throws SQLException
		{
			return rs.next() ? rs.getLong(1) : 0;
		}

		
		
	}
}
