/*
	$Id: FileSetMember.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.FileSetMemberData;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class represents a file that is member of a file set.

	@author Nicklas
	@version 2.5
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (må, 20 apr 2015) $
*/
public class FileSetMember
	extends BasicItem
{
	
	/**
		The type of item represented by this class.
		@see Item#FILESETMEMBER
		@see #getType()
	*/
	public static final Item TYPE = Item.FILESETMEMBER;

	/**
		This filter gives everybody read permission to file set members.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();

	
	/**
		Create a new member in a file set.
		@see FileSet#setMember(File, DataFileType)
	*/
	static FileSetMember getNew(DbControl dc, FileSet fileSet, File file, DataFileType type)
	{
		FileSetMember member = dc.newItem(FileSetMember.class);
		FileSetMemberData data = member.getData();
		data.setFileSet(fileSet.getData());
		data.setFile(file.getData());
		data.setDataFileType(type.getData());
		data.setValid(null);
		return member;
	}
	
	/**
		Get a <code>FileSetMember</code> object when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>FileSetMember</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the items
		@throws BaseException If there is another error
	*/
	public static FileSetMember getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		FileSetMember member = dc.loadItem(FileSetMember.class, id);
		if (member == null) throw new ItemNotFoundException("FileSetMember[id="+id+"]");
		return member;
	}
	
	/**
		Get a query configured to retrieve <code>FileSetMember</code> items.
		Note that the query can only retreive files for a single file set at
		a time. If fileSet==null when this method is called the client code must
		call {@link ItemQuery#setEntityParameter(String, BasicItem)} with
		a "fileSet" parameter before the query is executed.
		
		@param fileSet The file set to get members for (can be null)
		@return An {@link ItemQuery} object
		@throws InvalidUseOfNullException If required argument is null.
	 	@throws BaseException If there is another error.		
	*/
	public static ItemQuery<FileSetMember> getQuery(FileSet fileSet)
		throws BaseException
	{
		ItemQuery<FileSetMember> query = new ItemQuery<FileSetMember>(FileSetMember.class, RUNTIME_FILTER);
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property("fileSet"), 
					Hql.entityParameter("fileSet", Item.FILESET)
				)
			);
		if (fileSet != null)
		{
			query.setPermanentEntityParameter("fileSet", fileSet);
		}
		return query;
	}

	/**
		Creates a new member item from the given data.
		@param data the data.
	*/
	FileSetMember(FileSetMemberData data)
	{
		super(data);
	}
	
	@Override
	FileSetMemberData getData()
	{
		return (FileSetMemberData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
	From the BasicItem class
	-------------------------------------------
	*/
	/**
		READ permission is granted to all users. CREATE, WRITE and DELETE permissions are 
		granted if the logged in user has WRITE permission on the associated file set.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{	
		FileSet fileSet = getFileSet();
		if (fileSet.hasPermission(Permission.WRITE))
		{
			granted |= Permission.grant(Permission.CREATE, Permission.READ, Permission.WRITE, Permission.DELETE);				
		}
		else if (fileSet.hasPermission(Permission.USE))
		{
			granted |= Permission.grant(Permission.USE);
		}
		else if (fileSet.hasPermission(Permission.READ))
		{
			granted |= Permission.grant(Permission.READ);				
		}	
		super.initPermissions(granted, denied);
	}
	@Override
	PluginPermission getPluginPermissions()
	{
		return PluginPermission.NO_PLUGIN;
	}
	// -------------------------------------------
	
	/**
		Get the file set the file is a member of.
		@return A file set item
	*/
	public FileSet getFileSet()
	{
		return getDbControl().getItem(FileSet.class, getData().getFileSet());		
	}
	
	/**
		Get the file this item is representing in the file set.
		@return A file item
	*/
	public File getFile()
	{
		return getDbControl().getItem(File.class, getData().getFile());
	}
	
	/**
		Get the type of the file.
		@return A file set member type item
	*/
	public DataFileType getDataFileType()
	{
		return getDbControl().getItem(DataFileType.class, getData().getDataFileType());
	}
	
	/**
		Check if the file has been validated and if it passed or not.
		@return NULL if the file hasn't been validated, TRUE or FALSE 
			to indicate the result of the validation
	*/
	public Boolean isValid()
	{
		return getData().isValid();
	}
	
	/**
		Set the validation status of this member.
		@param valid If the member file is valid or not, use null to
			indicate that no validation has been performed
		@param errorMessage The error message if the validation failes
	*/
	public void setValid(Boolean valid, String errorMessage)
	{
		checkPermission(Permission.WRITE);
		getData().setValid(valid);
		getData().setErrorMessage(errorMessage);
	}
	
	/**
		Get the error message if the validation failed.
		@return The error message or null if the file is valid or if it
			hasn't been validated
	*/
	public String getErrorMessage()
	{
		return getData().getErrorMessage();
	}
	
	/**
		Checks if this file is part of the platform/variant the parent
		item belongs to.
		@return TRUE if the file is part of the platform, FALSE otherwise
		@since 2.10
	*/
	public boolean isPlatformFile()
	{
		FileStoreEnabled item = getFileSet().getItem();
		Platform platform = item.getPlatform();
		PlatformVariant variant = item.getVariant();
		if (platform == null) return false;
		return platform.getFileType(getDataFileType(), variant, false) != null;
	}

	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			// Do not add any filters.
		}
	}
}
