/**
	$Id: AnnotationSetLogger.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log.db;

import java.util.List;

import org.hibernate.Query;

import net.sf.basedb.core.Item;
import net.sf.basedb.core.LogControl;
import net.sf.basedb.core.PredefinedQuery;
import net.sf.basedb.core.data.AnnotationSetData;
import net.sf.basedb.core.data.ChangeHistoryDetailData;
import net.sf.basedb.core.data.LoggableData;
import net.sf.basedb.core.log.EntityDetails;
import net.sf.basedb.core.log.ChangeType;
import net.sf.basedb.core.log.EntityLogger;
import net.sf.basedb.core.log.LogManager;

/**
	Logger for annotations sets that detects and logs
	changes to inherited anntations as updates to the parent 
	item that are inheriting the annotations.
	
	@author Nicklas
	@version 2.13
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (må, 20 apr 2015) $
*/
public class AnnotationSetLogger
	implements EntityLogger
{

	private final DbLogManagerFactory factory;
	private final boolean detailedProperties;
	private final boolean oldPropertyValues;
	
	public AnnotationSetLogger(DbLogManagerFactory factory)
	{
		this.factory = factory;
		this.detailedProperties = factory.isLoggingDetailedProperties();
		this.oldPropertyValues = factory.isLoggingOldPropertyValues();
	}
	
	/*
		From the EntityLogger interface
		-------------------------------
	*/
	@Override
	public void logChanges(LogManager logManager, EntityDetails details)
	{
		// We are only logging update events. CREATE and DELETE events are
		// usually also catched by other loggers
		if (details.getChangeType() != ChangeType.UPDATE) return;
		
		LogControl logControl = logManager.getLogControl();
		AnnotationSetData as = (AnnotationSetData)details.getEntity();
		
		// Get the type of the parent item
		Item parentType = Item.fromValue(as.getItemType());
		
		Class<?> parentClass = parentType.getDataClass();
		// If the parent item isn't implementing the LoggableData interface
		// we don't log the changes
		if (!LoggableData.class.isAssignableFrom(parentClass)) return;
		
		// Get the id of the parent item 
		// NOTE! Do not use as.getItemId() since it will be obsolete if the item is deleted
		Query q = logControl.createHqlQuery(
			PredefinedQuery.getQueryString("DBLOG_GET_ITEMID_WITH_ANNOTATION", parentClass.getName()));
		q.setInteger("annotationSetId", as.getId());
		Integer parentId = (Integer)q.uniqueResult();
		
		// May happen in case the parent item has been deleted in this transaction
		if (parentId == null) return;
		
		// Generate list of changes
		DbLogManager dbLogManager = (DbLogManager)logManager;
		List<ChangeHistoryDetailData> changes = details.toChangeDetails(parentType, parentId, "annotationSet", 
			ChangeType.UPDATE, detailedProperties, oldPropertyValues);
		dbLogManager.logChangeDetails(changes, details);
	}
	// -------------------------------
}
