/**
	$Id: AbstractFilenameGenerator.java 5212 2010-01-12 13:21:49Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

import java.util.HashSet;
import java.util.Set;

/**
	Abstract superclass for file names generator implementions that will help
	with ensuring that a generator creates unique file names. Subclasses may simply
	call {@link #makeUnique(String)} to ensure that no duplicate file names are
	generated.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-01-12 14:21:49 +0100 (ti, 12 jan 2010) $
*/
public abstract class AbstractFilenameGenerator<T>
	implements FilenameGenerator<T>
{
	
	private Set<String> generatedFiles;
	
	protected AbstractFilenameGenerator()
	{
		generatedFiles = new HashSet<String>();
	}
	
	/**
		Checks if the suggested file name has already been generated by this
		file name generator or not.
		
		@param filename The filename to check
		@return TRUE if the name is unique or FALSE if it has already been used
	*/
	protected boolean isUnique(String filename)
	{
		return !generatedFiles.contains(filename);
	}
	
	/**
		Add the file name to the list of already generated file names.
		@param filename The file name to add
		@return TRUE if the file name has not been used before, FALSE otherwise
	*/
	protected boolean addFilename(String filename)
	{
		return generatedFiles.add(filename);
	}
	
	/**
		Make sure that a file name is unique. If the suggested file name is already
		unique the input parameter is returned. Otherwise a new filename is generated
		by appending a counter value to it until a unique name is found. The counter 
		value is added before any extension. Eg. example.txt, example-2.txt, example-3.txt
		would be generated if this method was called with "example.txt" as a parameter
		three times. The returned filename is automatically added to the list of known
		generated file names.
		
		@param filename The file name to check
		@return A file name that is sure to be unique
	*/
	protected String makeUnique(String filename)
	{
		int counter = 2;
		int dotIndex = filename.indexOf('.');
		String prefix = dotIndex > 0 ? filename.substring(0, dotIndex) : filename;
		String suffix = dotIndex > 0 ? filename.substring(dotIndex) : "";
		while (!isUnique(filename))
		{
			filename = prefix + "-" + counter + suffix;
			counter++;
		}
		addFilename(filename);
		return filename;
	}
	
	
}
