/**
	$Id: Node.java 6875 2015-04-20 09:08:18Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview;

import java.util.Collections;
import java.util.Comparator;
import java.util.IdentityHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.filter.IdentityFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;

/**
 	A node in an overview. There are two types of nodes: folder nodes
 	and item nodes. The only difference is that an item node has a link
 	to a {@link BasicItem} (see {@link #getItem()} but a folder node does not.
 	Both node types can have child nodes.
 	<p>
 	Each node has a unique ID ({@link #getId()}) which is based on the system 
 	hashcode. Each node also has a name which should be unique among the
 	child nodes. If the name isn't unique the {@link #getChild(String)} returns the
 	first matching node.
 	<p>
 	

	@author Nicklas
	@version 2.2
	@base.modified $Date: 2015-04-20 11:08:18 +0200 (må, 20 apr 2015) $
*/
public class Node
	implements Comparable<Node>
{
	private final String name;
	private final String title;
	private BasicItem item;
	private final Node parent;
	private final Type type;
	private final String nodeId;
	private final ChildNodeDirection direction;
	private List<Node> children;
	private int numWarnings;
	private int numErrors;
	private int childWarnings;
	private int childErrors;
	private boolean childrenLoaded;
	private Map<NodeAttribute<?>, Object> attributes;
	private List<Node> newChildren;
	
	/**
		Create an item-type root node.

		@param name The name of the node
		@param title The display title of the node
		@param item The item the node references
		@see GenericOverview#getRootNode()
	*/
	public Node(String name, String title, BasicItem item)
	{
		this(name, title, null, item, Type.ITEM, ChildNodeDirection.ALL);
	}
	
	/**
		Create a folder-type child node.
		@param name The name of the child node
		@param title The display title of the node
		@param parent The parent node
	*/
	public Node(String name, String title, Node parent)
	{
		this(name, title, parent, null, Type.FOLDER, ChildNodeDirection.NONE);
	}
	
	/**
		Create a folder-type child node.
		@param name The name of the child node
		@param title The display title of the node
		@param parent The parent node
		@param direction The direction to use when continuing to
			load child nodes
		@since 2.10
	*/
	public Node(String name, String title, Node parent, ChildNodeDirection direction)
	{
		this(name, title, parent, null, Type.FOLDER, direction);
	}
	
	/**
		Create an item-type child node
		@param name The name of the child node
		@param title The display title of the node
		@param parent The parent node
		@param item The item this node references
	 */
	public Node(String name, String title, Node parent, BasicItem item)
	{
		this(name, title, parent, item, Type.ITEM, ChildNodeDirection.NONE);
	}
	
	/**
		Create an item-type child node
		@param name The name of the child node
		@param title The display title of the node
		@param parent The parent node
		@param item The item this node references
		@param direction The direction to use when continuing to
			load child nodes
		@since 2.10
	*/
	public Node(String name, String title, Node parent, BasicItem item, ChildNodeDirection direction)
	{
		this(name, title, parent, item, Type.ITEM, direction);
	}
	
	private Node(String name, String title, Node parent, BasicItem item, Type type, ChildNodeDirection direction)
	{
		this.name = name;
		this.title = title;
		this.parent = parent;
		this.item = item;
		this.type = type;
		this.direction = direction == null ? ChildNodeDirection.NONE : direction;
		this.childrenLoaded = this.direction == ChildNodeDirection.NONE;
		this.nodeId = "N" + hashCode();
		if (parent != null)
		{
			parent.addChild(this);
		}
	}

	/**
		Make a copy of a node
		@since 2.10
	*/
	private Node(Node copyFrom, Node parent)
	{
		this(copyFrom.name, copyFrom.title, parent, copyFrom.item, copyFrom.type, copyFrom.direction);
	}
	
	
	/**
		Get a unique ID for this node. The ID is based on the hashcode and should be unique
		for a given virtual machine.
	*/
	public String getId()
	{
		return nodeId;
	}
	
	/**
		Get the node type of this node. Note that for folder-type nodes
		the {@link #getItem()} always return null. It it returns null for 
		an item-type node it usually means that access was denied to that
		particular item. 
	*/
	public Type getNodeType()
	{
		return type;
	}
	
	/**
		Get the name of this node. If the name is unique among the child
		nodes for a parent the call <code>getParent().getChild(this.getName())</code>
		should return this node.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Get the display title of this node.
	*/
	public String getTitle()
	{
		return title;
	}
	
	/**
		Get the item that is related to this node. Folder-type nodes always
		return null. If an item-type node returns null it usually means that
		access was denied to the item.
	*/
	public BasicItem getItem()
	{
		return item;
	}
	
	/**
		Get the item that is related to this node and make sure that it is
		loaded with the current DbControl. Folder-type nodes always
		return null. If an item-type node returns null it usually means that
		access was denied to the item.
		@since 2.10
	*/
	public BasicItem getItem(DbControl dc)
	{
		if (item != null)
		{
			item = item.getType().getById(dc, item.getId());
		}
		return item;
	}
	
	/**
		Get the item type of the item that this node related to this node.
		This method is a shortcut for <code>getItem().getType()</code>.
		@return The item type, or null if no item is attached to this
			node
		@since 2.10
	*/
	public Item getItemType()
	{
		return item == null ? null : item.getType();
	}
	
	/**
		Get the parent of this node. Returns null if this is the root node.
	*/
	public Node getParent()
	{
		return parent;
	}
	
	/**
		Find the first parent node that matches a filter. Eg.
		The first parent were {@link Filter#evaluate(Object)}
		returns true is the node that is returned.
		
		@param filter The filter to use for matching parent
		@return The node, or null if no node was found
		@since 2.10
	*/
	public Node getFirstParent(Filter<? super Node> filter)
	{
		Node parent = getParent();
		while (parent != null)
		{
			if (filter.evaluate(parent)) return parent;
			parent = parent.getParent();
		}
		return null;
	}
	
	/**
		Find the first node (including this) that matches
		a filter. Eg. the first node were {@link Filter#evaluate(Object)}
		returns true is the node that is returned. Calling
		this method is equivalent to first evaluating the
		filter on this node, and then calling
		{@link #getFirstParent(Filter)} if it didn't match.
		
		@param filter The filter to use for matching nodes
		@return The node, or null if no node was found
		@since 3.0
	*/
	public Node getFirstNode(Filter<? super Node> filter)
	{
		return filter.evaluate(this) ? this : getFirstParent(filter);
	}
	
	/**
		Find all nodes in the parent chain that matches the given
		filter.
		@param filter The filter to use for matching nodes
		@return A list with the matched nodes in the order they matched
		@since 3.2
	*/
	public List<Node> findAll(Filter<? super Node> filter)
	{
		List<Node> result = new LinkedList<Node>();
		Node current = this;
		while (current != null)
		{
			if (filter.evaluate(current)) result.add(current);
			current = current.getParent();
		}
		return result;
	}
	
	/**
		Get all child nodes of this node. May return null or an empty
		list if this node has no children.
	*/
	public List<Node> getChildren()
	{
		return children;
	}
	
	/**
		Get the number of child nodes in this node.
		@return The number of child nodes, or -1 if the child nodes
			has not been loaded yet
		@since 2.10
	*/
	public int numChildren()
	{
		if (!childrenLoaded) return -1;
		return children == null ? 0 : children.size();
	}
	
	/**
		Get the child node with a specific name. If more than one
		node has the same name only the first one found is returned.
		@param name The name of the child node.
		@return The child node with the given name, or null if no node is found
	*/
	public Node getChild(String name)
	{
		if (children != null)
		{
			for (Node child : children)
			{
				if (name.equals(child.getName()))
				{
					return child;
				}
			}
		}
		return null;
	}
	
	/**
		Get the number of new children added to this node
		since the last call to {@link #getNewChildren()}.
		@return The number of child nodes, or -1 if the child nodes
			has not been loaded yet
		@since 3.2
	*/
	public int numNewChildren()
	{
		if (!childrenLoaded) return -1;
		return newChildren == null ? 0 : newChildren.size();
	}
	
	/**
		Get all new children to this node. May return null or an empty
		list if this node has no children. Calling this method clears
		the new children information and the next call will return null
		unless more children is added inbetween.
		@since 3.2
	*/
	public List<Node> getNewChildren()
	{
		List<Node> result = newChildren;
		newChildren = null;
		return result;
	}
	
	/**
		Add a child node to this one.
	*/
	protected void addChild(Node child)
	{
		if (children == null) children = new LinkedList<Node>();
		if (newChildren == null) newChildren = new LinkedList<Node>();
		children.add(child);
		newChildren.add(child);
		this.childrenLoaded = true;
	}
	
	/**
		Sorts the list of child nodes.
		@param comparator A comparator that determines the sort order, 
			if not specifed, they children are sorted according to their
			title
		@since 2.14
	 */
	public void sortChildren(Comparator<? super Node> comparator)
	{
		if (children == null) return;
		Collections.sort(children, comparator);
	}
	
	/**
		Sets a flag to indicate that all children has been loaded.
		@since 2.10
	*/
	public void setChildrenLoaded()
	{
		this.childrenLoaded = true;
	}
	
	/**
		@return TRUE if the children to this node has been loaded, FALSE otherwise
		@since 2.10
	*/
	public boolean isChildrenLoaded()
	{
		return childrenLoaded;
	}

	/**
		Get the direction to use when loading children of this node.
		@return A ChildNodeDirection object
		@since 2.10
	*/
	public ChildNodeDirection getChildNodeDirection()
	{
		return direction;
	}
	
	/**
		Set the number of warnings that was found while validating this
		node. This method automatically updates the number of child
		warnings reported by parent nodes.
		@param numWarnings The number of warnings
		@see #addWarnings(int)
	*/
	public void setWarnings(int numWarnings)
	{
		if (parent != null) parent.addChildWarnings(numWarnings-this.numWarnings);
		this.numWarnings = numWarnings;
	}
	
	/**
		Add more warnings to this node. This method automatically updates the 
		number of child warnings reported by parent nodes.
		@param numWarnings The numer of warnings to add
		@see #setWarnings(int)
	*/
	public void addWarnings(int numWarnings)
	{
		this.numWarnings += numWarnings;
		if (parent != null) parent.addChildWarnings(numWarnings);
	}
	
	/**
		Add child warnings to this node. This method is called by child 
		nodes when the warnings are updated on them.
	*/
	protected void addChildWarnings(int numWarnings)
	{
		this.childWarnings += numWarnings;
		if (parent != null) parent.addChildWarnings(numWarnings);
	}
	
	/**
		Set the number of errors that was found while validating this
		node. This method automatically updates the number of child
		errors reported by parent nodes.
		@param numErrors The number of errors
		@see #addErrors(int)
	*/
	public void setErrors(int numErrors)
	{
		if (parent != null) parent.addChildErrors(numErrors-this.numErrors);
		this.numErrors = numErrors;
	}
	
	/**
		Add more errors to this node. This method automatically updates the 
		number of child errors reported by parent nodes.
		@param numErrors The numer of errors to add
		@see #setErrors(int)
	*/
	public void addErrors(int numErrors)
	{
		this.numErrors += numErrors;
		if (parent != null) parent.addChildErrors(numErrors);
	}
	
	/**
		Add child errors to this node. This method is called by child 
		nodes when the errors are updated on them.
	*/
	protected void addChildErrors(int numErrors)
	{
		this.childErrors += numErrors;
		if (parent != null) parent.addChildErrors(numErrors);
	}
	
	/**
		Get the number of warnings on this node.
		@see #getChildWarnings()
	*/
	public int getNumWarnings()
	{
		return numWarnings;
	}
	
	/**
		Get the number of errors on this node.
		@see #getChildErrors()
	*/
	public int getNumErrors()
	{
		return numErrors;
	}
	
	/**
		Get the number of warnings to children of this node.
		@see #getNumWarnings()
	*/
	public int getChildWarnings()
	{
		return childWarnings;
	}
	
	/**
		Get the number of errors to children of this node.
		@see #getNumErrors()
	*/
	public int getChildErrors()
	{
		return childErrors;
	}
	
	/**
		Clear all warnings, errors on this node and on all child nodes.
	*/
	public void clearFailures()
	{
		numErrors = 0;
		numWarnings = 0;
		childWarnings = 0;
		childErrors = 0;
		if (getChildren() != null)
		{
			for (Node child : getChildren())
			{
				child.clearFailures();
			}
		}
	}

	/**
		Set a node attribute.
		@param attribute The attribute to set
		@param value The value to set for the attribute or null to remove the
			current value
		@since 3.0
	*/
	public <T> void setAttribute(NodeAttribute<T> attribute, T value)
	{
		if (value == null)
		{
			if (attributes != null)	attributes.remove(attribute);
		}
		else
		{
			if (attributes == null) attributes = new IdentityHashMap<NodeAttribute<?>, Object>();
			attributes.put(attribute, value);
		}
	}
	
	/**
		Get the value for a node attribute.
		@param attribute The attribute to get
		@return The value or null if there is no value
		@since 3.0
	*/
	@SuppressWarnings("unchecked")
	public <T> T getAttribute(NodeAttribute<T> attribute)
	{
		T value = null;
		if (attributes != null) value = (T)attributes.get(attribute);
		return value;
	}
	
	/**
		Check if the node has a value for the given attribute.
		@param attribute The attribute to check
		@return TRUE if a value exists, FALSE otherwise
		@since 3.0
	*/
	public boolean hasAttribute(NodeAttribute<?> attribute)
	{
		return attributes != null && attributes.containsKey(attribute);
	}
	
	/**
		Create a deep copy of this node and all it's children. The copy
		will be attached to the given parent.
		@param parentOfCopy The node that should be the parent of the
			copy or null to make the copy a root node
		@return The copy
		@throws PermissionDeniedException If the given parent node is a
			descendant of this node (it would lead to an infinite loop)
		@since 2.10
	*/
	public Node deepCopy(Node parentOfCopy)
	{
		if (parentOfCopy.getFirstParent(new IdentityFilter<Node>(this)) != null)
		{
			throw new PermissionDeniedException("The parent node '" + parentOfCopy.getName() + 
				"' is a descendant of this node: " + getName());
		}
		return deepCopyInternal(parentOfCopy);
	}
	
	private Node deepCopyInternal(Node parentOfCopy)
	{
		Node copy = new Node(this, parentOfCopy);
		copy.childrenLoaded = this.childrenLoaded;
		List<Node> children = getChildren();
		if (children != null)
		{
			for (Node child : children)
			{
				Node childCopy = child.deepCopyInternal(copy);
			}
		}
		return copy;
	}
	
	@Override
	public int compareTo(Node o)
	{
		int tmp = title.compareTo(o.title);
		return tmp != 0 ? tmp : nodeId.compareTo(o.nodeId);
	}

	
	@Override
	public String toString()
	{
		return "Node[" + type + "; id=" + nodeId + "; name=" + name + "; title=" + title + "]";
	}
	
	/**
		The type of node.
	*/
	public enum Type
	{
		FOLDER, ITEM;
	}

}
