/**
	$I$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;


import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RootRawBioAssay;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.NodeAttribute;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for root raw bioassays. Used in the 
	reverse-loading direction only when going from experiments
	to raw bioassays. This loader replaces the raw bioassays
	loader.

	@author Nicklas
	@since 3.6
*/
public class RootRawBioAssayLoader
	extends BasicItemNodeLoader<RootRawBioAssay>
{
	public RootRawBioAssayLoader()
	{
		super(Item.ROOTRAWBIOASSAY, DENY_ROOT_NODE, 
				new NameableNameGenerator<RootRawBioAssay>("rawbioassay", "Raw bioassay"));
	}

	/*
		From the NodeLoader interface
		------------------------------
	*/

	/**
		Create a reverse-loading raw bioassay nodes from a given experiment node.
		The returned node is a folder-type node containing nodes for each raw
		bioassay is part of the experiment.
		@return The folder node for the raw bioassays
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node experimentNode)
	{
		NodeFactory<RootRawBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		Experiment experiment = (Experiment)experimentNode.getItem(dc);
		ItemQuery<RootRawBioAssay> query = context.initQuery(experiment.getRootRawBioAssays(), "name");
		ItemResultIterator<RootRawBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			RootRawBioAssay root = it.next();
			RawBioAssay raw = root.getRawBioAssay();
			if (folderNode == null)
			{
				folderNode =  new Node("rawbioassays", "Raw bioassays", experimentNode, ChildNodeDirection.REVERSE);
			}
			Node rawNode = createItemNode(nf, root, root, false, folderNode, ChildNodeDirection.REVERSE);
			try
			{
				// We must set the EXTRACT attribute so that upstream loaders
				// follow the correct path when reaching the biomaterials section
				rawNode.setAttribute(NodeAttribute.EXTRACT, raw.getParentExtract());
			}
			catch (PermissionDeniedException ex)
			{}
		}
		postValidateFolder(nf, folderNode, experimentNode, false);
		return folderNode;
	}
	// ------------------------------
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Load property nodes for a raw bioassay:
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Data files: {@link DataFileLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Platform: {@link PlatformLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Software: {@link SoftwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Array design: {@link ArrayDesignLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.EXTRACT).createReverseNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.FILESETMEMBER).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.PLATFORM).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ARRAYDESIGN).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.SOFTWARE).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, rawBioAssayNode);
	}
	/**
		Loads the derived bioassay node that this raw bioassay is created from.
		@see DerivedBioAssayLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.DERIVEDBIOASSAY).createReverseNode(dc, context, rawBioAssayNode);
	}
	// --------------------------------

}
