/*
	$Id: TestExtract.java 6775 2015-03-17 12:40:32Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.ResultList;
import net.sf.basedb.util.listable.ExtractToChildExtractTransformer;
import net.sf.basedb.util.listable.ExtractToParentExtractTransformer;
import net.sf.basedb.util.listable.ExtractToSampleTransformer;
import net.sf.basedb.util.listable.SampleToExtractTransformer;
import net.sf.basedb.util.listable.SourceItemTransformer;
import net.sf.basedb.util.listable.TransformContext;

import java.util.Collections;
import java.util.Date;
import java.util.Set;

public class TestExtract
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing extracts");
		write_header();
		// Standard tests: create, load, list
		int sampleId = TestSample.test_create(0, "Parent sample", 0, true);
		int tagId = TestTag.test_create("Test", SystemItems.getId(Tag.LABEL), true);
		int id = test_create(0, "Extract #1 (standalone)", 0, 0, true);
		int id2 = test_create(sampleId, "Extract #2 (sample parent)", 0, 0, false);
		int id3 = test_create(id, "Extract #3 (extract parent+labeled)", SystemItems.getId(Extract.LABELED), tagId, false);
		int id4 = test_create_pooled("Extract #4 (pooled)", id, id2, id3);

	
		test_load(id);
		test_list(0, -1);
		
		write_events_header();
		test_add_event(id, 100.0f);
		test_list_events(id, 4);  // creation event, create event for id3+id4, and extra event
		
		write_parents_header();
		test_list_sources(id4, 3); // The three id, id2 and id3 samples
		test_list_children(id, 2); // id3 and id4

		// Transformers
		test_childrenTransformer(id, 2);
		test_parentsTransformer(id4, 3);
		test_extractToSampleTransformer(id2, 1);
		test_sampleToExtractTransformer(sampleId, 1);

		
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		// Standard test: Delete
		test_delete(id4);
		test_delete(id3);
		test_delete(id2);
		test_delete(id);
		TestSample.test_delete(sampleId);
	
		write("++Testing extracts "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int parentId, String name, int subtypeId, int tagId, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.EXTRACT)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getNew(dc);
			e.setName(name);
			if (setAll)
			{
				e.setDescription("Added at "+new Date());
				e.setExternalId("XCV-455-EXTRACT");
				e.setOriginalQuantity(500.0f);
			}
			if (parentId != 0)
			{
				BioMaterial parent = BioMaterial.getById(dc, parentId);
				BioMaterialEventSource evtSrc = e.getCreationEvent().setSource(parent);
				evtSrc.setUsedQuantity(200.0f);
			}
			if (subtypeId > 0)
			{
				e.setItemSubtype(ItemSubtype.getById(dc, subtypeId));
			}
			if (tagId > 0)
			{
				e.setTag(Tag.getById(dc, tagId));
			}
			dc.saveItem(e);
			dc.commit();
			id = e.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(e, false);
			write_item(0, e);
			write("--Create extract OK");
		}
		catch (Throwable ex)
		{
			write("--Create extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static int test_create_pooled(String name, int... sourceIds)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.EXTRACT)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getNew(dc);
			e.setName(name);
			e.setDescription("Added at "+new Date());
			e.setExternalId("POOLED-455-EXTRACT");
			e.setOriginalQuantity(500.0f);
			BioMaterialEvent evt = e.getCreationEvent();
			for (int sourceId : sourceIds)
			{
				Extract s = Extract.getById(dc, sourceId);
				Float original = s.getOriginalQuantity();
				BioMaterialEventSource eventSource = evt.addSource(s);
				if (original != null)
				{
					eventSource.setUsedQuantity(original/2);
				}
			}
			dc.saveItem(e);
			dc.commit();
			id = e.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(e, false);
			write_item(0, e);
			write("--Create pooled extract OK");
		}
		catch (Throwable ex)
		{
			write("--Create pooled extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}
	
	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, id);
			write_item(0, e);
			write("--Load extract OK");
		}
		catch (Throwable ex)
		{
			write("--Load extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static boolean test_list(int subtypeId, int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemQuery<Extract> query = Extract.getQuery();
			if (subtypeId > 0)
			{
				query.restrict(Restrictions.eq(Hql.property("itemSubtype"), Expressions.integer(subtypeId)));
			}
			ItemResultList<Extract> l = query.list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List extracts OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List extracts FAILED");
			ex.printStackTrace();
			ok = false;
			return false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return true;
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, id);
			dc.deleteItem(e);
			Set<ItemProxy> using = e.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(e + " is used by " + using);
			}
			dc.commit();
			write("--Delete extract OK");
		}
		catch (Throwable ex)
		{
			write("--Delete extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Extract> l = Extract.getQuery().list(dc);
			int numDeleted = Trashcan.delete(dc.getSessionControl(), l, true, null);
			if (numDeleted != l.size()) throw new BaseException("Could not delete all items: " + numDeleted + " of " + l.size());
			dc.commit();
			write("--Delete all extracts OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all extracts FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tSubtype\tDescription\tExternal id\tParent type\tParent\tTag\tQuantity\tRemain");
			write("-- \t-- \t--------- \t-------\t-----------\t-----------\t-----------\t------\t---\t--------\t------");
		}
	}
	static void write_item(int i, Extract e)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+e.getId()+"\t"+e.getName()+"\t"+e.getItemSubtype()+"\t"+e.getDescription()+
			"\t"+e.getExternalId()+"\t"+e.getParentType()+"\t"+e.getParent()+"\t"+e.getTag()+"\t"+e.getOriginalQuantity()+"\t"+e.getRemainingQuantity());
	}
	static void write_events_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tBiomaterial\tEvent type\tEvent date\tEntry date\tUsed quantity\tComment");
			write("-- \t-- \t-----------\t----------\t----------\t----------\t-------------\t-------");
		}
	}
	static void write_item(int i, BioMaterialEvent evt, Extract e)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+evt.getId()+"\t"+evt.getBioMaterial()+"\t"+evt.getEventType()+
			"\t"+evt.getEventDate()+"\t"+evt.getEntryDate()+"\t"+evt.getUsedQuantity(e)+"\t"+evt.getComment());
	}
	static void write_item(int i, PhysicalBioAssay h)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+h.getId()+"\t"+h.getName()+"\t"+h.getDescription());
	}

	static void write_parents_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tExternal id\tQuantity\tRemain\tUsed");
			write("-- \t-- \t--------- \t-----------\t-----------\t--------\t------\t----");
		}
	}

	static void write_item(int i, MeasuredBioMaterial b, BioMaterialEvent evt)
		throws BaseException
	{
		if (!TestUtil.getSilent()) System.out.println(i+":\t"+b.getId()+"\t"+b.getName()+"\t"+b.getDescription()+
			"\t"+b.getExternalId()+"\t"+b.getParentType()+"\t"+b.getOriginalQuantity()+"\t"+b.getRemainingQuantity()+
			"\t"+evt.getUsedQuantity(b));
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	
	static int test_add_event(int extractId, float usedQuantity)
	{
		if (extractId == 0) return 0;
		int eventId = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			BioMaterialEvent bme = e.newEvent();
			bme.setUsedQuantity(usedQuantity);
			bme.setComment("Added a new event");
			dc.saveItem(bme);
			dc.commit();
			eventId = bme.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(bme, false);
			write_item(0, bme, e);
			write("--Add event for extract OK");
		}
		catch (Throwable ex)
		{
			write("--Add event for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return eventId;
	}
	
	static void test_list_events(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			ItemResultList<BioMaterialEvent> l = e.getEvents().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i), e);
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List events for extract OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List events for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_sources(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			BioMaterialEvent evt = e.getCreationEvent();
			ResultList<BioMaterialEventSource> l = evt.getEventSources().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				TestSample.write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List sources for extract OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List sources for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list_children(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			ItemResultList<Extract> ev = e.getChildExtracts().list(dc);
			
			for (int i=0; i<ev.size(); i++)
			{
				write_item(i, ev.get(i));
			}
			if (expectedResults >= 0 && expectedResults != ev.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+ev.size());
			}
			write("--List children for extract OK ("+ev.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List children for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	static void test_list_bioassays(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract le = Extract.getById(dc, extractId);
			ItemResultList<PhysicalBioAssay> l = le.getPhysicalBioAssays().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List bioassays for extract OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List bioassays for extract FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	
	static void test_childrenTransformer(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			SourceItemTransformer transformer = new ExtractToChildExtractTransformer(false);
			TransformContext tfc = new TransformContext(dc);
			Set<Integer> children = transformer.transform(tfc, Collections.singleton(e.getId()));

			int i = 0;
			for (int childId : children)
			{
				write_item(i, Extract.getById(dc, childId));
				i++;
			}
			if (expectedResults >= 0 && expectedResults != children.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+children.size());
			}
			write("--List children for extract using transformer OK ("+children.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List children for extract using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_parentsTransformer(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);

			SourceItemTransformer transformer = new ExtractToParentExtractTransformer(false);
			TransformContext tfc = new TransformContext(dc);
			Set<Integer> parents = transformer.transform(tfc, Collections.singleton(e.getId()));

			int i = 0;
			for (int parentId : parents)
			{
				write_item(i, Extract.getById(dc, parentId));
				i++;
			}
			if (expectedResults >= 0 && expectedResults != parents.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+parents.size());
			}
			write("--List parents for extract using transformer OK ("+parents.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List parents for extract using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_extractToSampleTransformer(int extractId, int expectedResults)
	{
		if (extractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Extract e = Extract.getById(dc, extractId);
			
			SourceItemTransformer transformer = new ExtractToSampleTransformer();
			TransformContext tfc = new TransformContext(dc);
			Set<Integer> parents = transformer.transform(tfc, Collections.singleton(e.getId()));

			int i = 0;
			for (int parentId : parents)
			{
				TestSample.write_item(i, Sample.getById(dc, parentId));
				i++;
			}
			if (expectedResults >= 0 && expectedResults != parents.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+parents.size());
			}
			write("--List sample for extract using transformer OK ("+parents.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List sample for extract using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_sampleToExtractTransformer(int sampleId, int expectedResults)
	{
		if (sampleId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Sample s = Sample.getById(dc, sampleId);
			
			SourceItemTransformer transformer = new SampleToExtractTransformer();
			TransformContext tfc = new TransformContext(dc);
			Set<Integer> children = transformer.transform(tfc, Collections.singleton(s.getId()));

			int i = 0;
			for (int childId : children)
			{
				write_item(i, Extract.getById(dc, childId));
				i++;
			}
			if (expectedResults >= 0 && expectedResults != children.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+children.size());
			}
			write("--List extracts for sample using transformer OK ("+children.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List extracts for sample using transformer FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	
}
